from vector import Vector
from quaternion import Quaternion

DAMPING = 0.01
ROTATION_DAMPING = 0.03

class Body:
    def __init__(self):

        # propeties
        self.mass = 1.0
        # assume a cube rotated around its axis for now, intertia really must be
        # a vector though
        self.inertia = 10 * self.mass / 6.0

        # state
        self.position = Vector(0, 0, 0)
        self.momentum = Vector(0, 0, 0)
        self.orientation = Quaternion(1, 0, 0, 0)
        self.angular_momentum = Vector(0, 0, 0)

        # derived
        self.velocity = Vector(0, 0, 0)
        self.spin = Quaternion(0, 0, 0, 0)
        self.r = Vector(1, 0, 0)
        self.u = Vector(0, 1, 0)
        self.b = Vector(0, 0, 1)

        # temporary
        self.force = Vector(0, 0, 0)
        self.torque = Vector(0, 0, 0)

    def move(self, sideway, upward, backward):
        self.force += self.r * sideway
        self.force += self.u * upward
        self.force += self.b * backward

    def turn(self, pitch, yaw, roll):
        self.torque += self.r * pitch
        self.torque += self.u * yaw
        self.torque += self.b * roll
    
    def velocity_at(self, rel):
        return self.velocity + self.angular_velocity.cross(rel)

    def tick(self):
        # apply velocity
        self.position += self.velocity
        # apply force
        self.momentum += self.force
        # apply spin
        self.orientation += self.spin
        # apply torque
        self.angular_momentum += self.torque
        # derive velocity
        self.velocity = self.momentum / self.mass
        # derive angular velocity
        self.angular_velocity = self.angular_momentum / self.inertia
        # re-normalize
        self.orientation.normalize()
        # get spin from angular velocity
        self.spin = self.angular_velocity.quatmul(self.orientation)
        self.spin *= 0.5
        # get orientation vectors
        self.r, self.u, self.b = self.orientation.vectors()
        # reset force and torque
        self.force = Vector(0, 0, 0)
        self.torque = Vector(0, 0, 0)

    def damp(self):
        # linear damping
        v = self.momentum.norm()
        self.force -= self.momentum * DAMPING
        # rotational damping
        v = self.angular_momentum.norm()
        self.torque -= self.angular_momentum * ROTATION_DAMPING

    def matrix(self):
        matrix = []
        matrix.append(self.r.x)
        matrix.append(self.r.y)
        matrix.append(self.r.z)
        matrix.append(0)
        matrix.append(self.u.x)
        matrix.append(self.u.y)
        matrix.append(self.u.z)
        matrix.append(0)
        matrix.append(self.b.x)
        matrix.append(self.b.y)
        matrix.append(self.b.z)
        matrix.append(0)
        matrix.append(self.position.x)
        matrix.append(self.position.y)
        matrix.append(self.position.z)
        matrix.append(1)

        return matrix
