import math, vector

class Quaternion:
    """
    A quaternion can be useful to hold rotations, as it is much easier to use
    for physics code. For example a 3x3 matrix has too much redundancy (9 values
    instead of 3), and so usually will be hard to deal with or lead to wrong
    results. A quaternion instead will hold not much more than just the
    orientation (4 values instead of 3).

    Orientation basically is a direction, with an additional rotation. For
    example, a 3d vector plus an angle would describe it. Since the length of
    the 3d vector does not matter, just 3 angles also are enough (pitch, yaw,
    roll). Obviously, we can convert from any form to each other, quaternions
    just have properties which make them easy to use in physics code.
    """
    def __init__(q, w, x, y, z):
        q.w = w
        q.x = x
        q.y = y
        q.z = z

    def __iadd__(q, p):
        q.w += p.w
        q.x += p.x
        q.y += p.y
        q.z += p.z
        return q

    def __imul__(q, s):
        q.w *= s
        q.x *= s
        q.y *= s
        q.z *= s
        return q

    def vectors(q):
        """
        Output three orientation vectors for the quaternion. That is, if the
        quaternion is used as a 3D orientation, return right/up/back vectors
        representing the same orientation.
        """
        ww = q.w * q.w
        xx = q.x * q.x
        yy = q.y * q.y
        zz = q.z * q.z
        wx = q.w * q.x * 2
        wy = q.w * q.y * 2
        wz = q.w * q.z * 2
        xy = q.x * q.y * 2
        xz = q.x * q.z * 2
        yz = q.y * q.z * 2

        rx = ww + xx - yy - zz
        ux = xy - wz
        bx = xz + wy

        ry = xy + wz
        uy = ww - xx + yy - zz
        by = yz - wx

        rz = xz - wy
        uz = yz + wx
        bz = ww - xx - yy + zz

        r = vector.Vector(rx, ry, rz)
        u = vector.Vector(ux, uy, uz)
        b = vector.Vector(bx, by, bz)

        return r, u, b

    def normalize(q):
        """
        Normalize the quaternion. This may be useful to prevent deteriorating
        the quaternion if it is used for a long time, due to floating point
        inaccuracies.
        """
        n = math.sqrt(q.w * q.w + q.x * q.x + q.y * q.y + q.z * q.z)
        q.w /= n
        q.x /= n
        q.y /= n
        q.z /= n

